/**
* Klasse, die einen Dateimanager verwaltet
*/
function FileManager(
    nodeParent,     // Element, als dessen Kind der Inhalt des Dateimanager erstellt werden soll
    boolClipboard   // Darf der Benutzer die Zwischenablage verwenden?
) {
    var doc = nodeParent.ownerDocument;
    this._nodeUl = doc.createElement('ul');
    this._nodeUl.setAttribute('class', 'filemanager');
    nodeParent.appendChild(this._nodeUl);
    Object.defineProperty(this, 'useClipboard', { enumerable: true, value: boolClipboard ? true : false, writeable: false });
    this._arrLis = [];
    this._root = new Anyfile(this);
    this._root.isDir = true;
    this._root.filename = '';
    this._current = this._root;

    var self = this;
    this._objBtnBar = new BtnBar(nodeParent);
    this._objBtnBar.add([
        { label: this._current.getPath() + '/', className: 'btn_home', title: 'Pfad ändern', alwaysShowLabel: true, onclick: function() {
            var objBubble = self._objBtnBar.showBubble();
            var submenu = new BtnBar(objBubble.node, true, objBubble, true);
            submenu.add([
                { label: 'Übergeordneter Ordner', className: 'btn_up', title: 'Zum übergeordneten Ordner wechseln', disabled: (self._current == self._root), onclick: function() {
                    if (self._current.parentDir instanceof Anyfile)
                        self.open(self._current.parentDir);
                } },
                { label: 'Wurzelordner', className: 'btn_home', title: 'Zum Wurzelordner gehen', disabled: (self._current == self._root), onclick: function() {
                    self.open(self._root);
                } },
                { label: 'Menü schließen', className: 'btn_close', title: 'Dieses Menü schließen' },
                true
            ]);
            var i = new ValidatedTextForm(objBubble.node, {
                label: 'Aktueller Pfad:',
                placeholder: 'Pfad',
                submitbtn: 'Wechseln',
                onsubmit(path) {
                    self.path = path;
                    self._objBtnBar.closeBubble();
                },
                onvalidate(path) {
                    var file = self._current.getByPath(path);
                    this.strMessage = file ? (file.isDir ? 'Ordner gefunden' : 'Das ist kein Ordner') : 'Pfad ist ungültig oder Datei nicht vorhandne';
                    return (file && file.isDir) ? true : false;
                },
                text: self.path
            });
            if (self.boolClickToOpen) {
                i.focus();
                i.select();
            }
        } },
        { label: 'Erstellen', className: 'btn_new', title: 'Neuen Unterordner oder neue Datei anlegen', alwaysShowLabel: true, onclick: function() {
            new NewAnyfileForm(self._objBtnBar, self);
        } }
    ]);
    if (this.useClipboard)
        this._objBtnBar.add([
            { label: 'Einfügen', className: 'btn_paste', title: 'Aus Zwischenablage einfügen', disabled: true, onclick: function() {
                if (!self._current.canInsert(self.clipboard))
                    return;
                var file = self.clipboard;
                if (self.boolMove) {
                    if (self._current.isIn(self.clipboard))
                        return;
                    self.clipboard.parentDir = self._current;
                    self.clipboard = null;
                } else {
                    self.clipboard.clone(true).parentDir = self._current;
                }
                self.selectFilename(file.filename);
            } }
        ]);
    this._objBtnBar.add([
        { label: 'Einstellungen', className: 'btn_preferences', title: 'Einstellungen ändern', onclick: function() {
            return [
                { label: 'Ordner zuerst anzeigen', className: self.boolDirsFirst ? 'btn_checked' : 'btn_notchecked', selected: self.boolDirsFirst, onclick: function() { self.boolDirsFirst = !self.boolDirsFirst; } },
                { label: 'Durch Anklicken öffnen', className: self.boolClickToOpen ? 'btn_checked' : 'btn_notchecked', selected: self.boolClickToOpen, onclick: function() { self.boolClickToOpen = !self.boolClickToOpen; } },
                { label: 'Menü schließen', className: 'btn_close', title: 'Dieses Menü schließen' }
            ];
        } }
    ]);
    var fktTouch = function() {
        self.boolClickToOpen = false;
        document.body.removeEventListener('touchstart', fktTouch);
    }
    document.body.addEventListener('touchstart', fktTouch);
}

// UL-Element zur Anzeige von Dateien
FileManager.prototype._nodeUl = null;

// Paare aus Nodes und zugehöriger Anyfile-Instanz
FileManager.prototype._arrLis = [];

// Wurzelordner
FileManager.prototype._root = null;

// Aktuell angezeigter Knoten
FileManager.prototype._current = null;

// Durch Anklicken öffnen?
FileManager.prototype.boolClickToOpen = true;

// Wird beim Einfügen des Elements aus der Zwischenablage verschoben statt kopiert?
FileManager.prototype.boolMove = false;

// Welche Anyfile-Instanz befindet sich aktuell in der Zwischenablage?
FileManager.prototype._clipboard = null;
Object.defineProperty(FileManager.prototype, 'clipboard', {
    get() { return this._clipboard; },
    set(obj) {
        this._clipboard = obj;
        this.reportChange();
    }
});

// Sollen Ordner zuerst angezeigt werden?
FileManager.prototype._boolDirsFirst = true;
Object.defineProperty(FileManager.prototype, 'boolDirsFirst', {
    get() { return this._boolDirsFirst; },
    set(bool) {
        if (bool == this._boolDirsFirst)
            return;
        this._boolDirsFirst = bool;
        this.reportChange();
    }
});

// Funktion, die bei jeder Änderung ausgeführt werden soll
FileManager.prototype.onchange = null;

// Funktion, die beim Öffnen einer Datei ausgeführt werden soll
FileManager.prototype.onopen = null;

// Zeitweise Änderungen ignorieren
FileManager.prototype._ignoreChanges = false;

/**
* Auf Änderung reagieren
* @return undefined
*/
FileManager.prototype.reportChange = function(
    file    // Anyfile-Instanz, die sich geändert hat
) {
    if (this._ignoreChanges)
        return;
    if (typeof this.onchange == 'function')
        this.onchange.call(this);
    if (this._objBtnBar) {
        var path = this.path;
        if (this._objBtnBar.items[0].label != path) {
            this._objBtnBar.items[0].label = path;
            this._objBtnBar.items[0].className = (this._current == this._root) ? 'btn_home' : 'btn_diropen';
        }
        // Zwischenablage
        if (this.useClipboard) {
            if (this._clipboard && !this._clipboard.isIn(this._root))
                this._clipboard = null;
            var mayPaste = (this._clipboard && this._current.canInsert(this._clipboard) && !(this.boolMove && this._current.isIn(this.clipboard))) ? true : false;
            if (this._objBtnBar.items[2].disabled == mayPaste)
                this._objBtnBar.items[2].disabled = !mayPaste;
        }
    }

    this.refresh();
}

// Aktueller Pfad
Object.defineProperty(FileManager.prototype, 'path', {
    get() {
        var s = this._current.getPath();
        return s.length ? s : '/';
    },
    set(s) {
        this.open(s);
    }
});

/**
* Für eine Datei zu verwendenden Klassenname bestimmen
* @return String    Klassenname
*/
FileManager.getFileClass = function(
    file        // Anyfile-Instanz, für die ein Klassenname gesucht wird
) {
    if (file.isDir)
        return 'dir';
    var type = file.type;
    if (!type)
        return '';
    switch (type) {
        case 'application/x-tar':
            return 'package';
        case 'application/pdf':
            return 'document';
        case 'executable':
        case 'application':
            return 'application';
        case 'text/html':
            return 'html';
    }
    if (type.indexOf('spreadsheet') != -1) return 'spreadsheet';
    if (type.indexOf('presentation') != -1) return 'presentation';
    if (type.indexOf('drawing') != -1) return 'drawing';
    if (type.indexOf('svg') != -1) return 'drawing';
    if (type.indexOf('document') != -1) return 'document';
    if (type.indexOf('office') != -1) return 'document';
    if (type.indexOf('zip') != -1) return 'package';
    switch (type.split('/')[0]) {
        case 'text':
            return 'text';
        case 'image':
            return 'image';
        case 'video':
            return 'video';
        case 'audio':
            return 'audio';
    }
    return '';
}

/**
* Datei file anzeigen
* @return undefined
*/
FileManager.prototype._showFile = function(
    file        // Anzuzeigende Anyfile-Instanz
) {
    var doc = this._nodeUl.ownerDocument;
    var li = doc.createElement('li');
    if (this.boolMove && this.useClipboard && (file == this._clipboard))
        li.className = 'cut';
    var a = doc.createElement('a');
    a.className = FileManager.getFileClass(file);
    a.appendChild(document.createTextNode(file.filename));
    a.href = '#';
    li.appendChild(a);
    this._nodeUl.appendChild(li);
    var self = this;
    this._arrLis.push({ file: file, node: li });
    a.addEventListener('click', function(ev) {
        ev.preventDefault();
        if (self._objBtnBar.activeBubble)
            return;
        ev.stopPropagation();
        if (self.boolClickToOpen)
            self.open(file);
        else
            self.contextmenu(file);
    });
    a.addEventListener('dblclick', function(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.contextmenu(file);
    });
    a.addEventListener('contextmenu', function(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.contextmenu(file);
    });
}

/**
* A-Element zur Anyfile-Instanz filename ermitteln
* @return AnchorElement     Zugehöriges A-Element
*/
FileManager.prototype._getElement = function(
    filename    // Dateiname der  Datei oder zugehörige Anyfile-Instanz
) {
    if (filename instanceof Anyfile) {
        if (filename.parentDir == this._current)
            filename = filename.filename;
        else
            return null;
    }
    for (var z = 0; z < this._arrLis.length; ++z) {
        if (this._arrLis[z].file.filename != filename)
            continue;
        return this._arrLis[z].node.firstChild;
    }
    return null;
}

/**
* Kontextmenü zur Anyfile-Instanz file anzeigen
* @return undefined
*/
FileManager.prototype.contextmenu = function(
    file        // Betroffene Anyfile-Instanz
) {
    var self = this;
    var objBubble = this._objBtnBar.showBubble();
    var submenu = new BtnBar(objBubble.node, true, objBubble, true);
    submenu.add([
        { label: 'Öffnen', className: 'btn_open', title: (file.isDir ? 'Diesen Ordner ' : 'Diese Datei ') + 'öffnen', onclick: function() {
            self.open(file);
        } }
    ]);
    if (this.useClipboard)
        submenu.add([
            { label: 'Kopieren', className: 'btn_copy', title: (file.isDir ? 'Diesen Ordner ' : 'Diese Datei ') + 'in die Zwischenablage kopieren', onclick: function() {
                self.boolMove = false;
                self.clipboard = file;
            } },
            { label: 'Ausschneiden', className: 'btn_cut', title: (file.isDir ? 'Diesen Ordner ' : 'Diese Datei ') + 'beim Einfügen verschieben', onclick: function() {
                self.boolMove = true;
                self.clipboard = file;
            } }
        ]);
    submenu.add([
        { label: 'Löschen', className: 'btn_delete', title: (file.isDir ? 'Diesen Ordner ' : 'Diese Datei ') + 'unwiederruflich löschen', onclick: function() {
            file.parentDir = null;
        } },
        { label: 'Zurück', className: 'btn_close', title: 'Dieses Menü schließen' },
        true
    ]);
    var i = new ValidatedTextForm(objBubble.node, {
        label: 'Umbenennen in:',
        submitbtn: 'Umbenennen',
        text: file.filename,
        onvalidate(filename) { return NewAnyfileForm.validateFilename(this, filename, self._current); },
        onsubmit(filename) {
            file.filename = filename;
            self._objBtnBar.closeBubble();
            self.selectFilename(file);
        }
    });
    if (self.boolClickToOpen) {
        i.focus();
        i.select();
    }
    objBubble.reposition(this._getElement(file));
}


/**
* Datei mit Namen filename markieren
* @return undefined
*/
FileManager.prototype.selectFilename = function(
    filename    // Dateiname der zu markierenden Datei oder zugehörige Anyfile-Instanz
) {
    var node = this._getElement(filename);
    if (node)
        node.focus();
}

/**
* Datei öffnen
* @return undefined
*/
FileManager.prototype.open = function(
    file        // Zu öffnende Anyfile-Instanz oder Pfad zu dieser
) {
    if (!(file instanceof Anyfile) && !((file = this._current.getByPath(file)) instanceof Anyfile))
        return;
    if (file.isDir) {
        this._current = file;
        this.reportChange();
    } else {
        if (file.parentDir && (file.parentDir != this._current))
            this.open(file.parentDir);
        var self = this;
        if ((typeof this.onopen != 'function') || !this.onopen.call(this, file)) {
            var objBubble = this._objBtnBar.showBubble();
            var submenu = new BtnBar(objBubble.node, true, objBubble, true);
            submenu.add([
                'Datei kann nicht geöffnet werden…',
                { label: 'Zurück', className: 'btn_close', title: 'Dieses Menü schließen' },
            ]);
            objBubble.reposition(this._getElement(file));
        }
    }
    /*!*/
}

/**
* Status und Inhalt aus JSON wiederherstellen
* @return undefined
*/
FileManager.prototype.fromJSON = function(
    json            // JSON, welches zu erstellende Instanz und Inhalt repräsentiert
) {
    this._ignoreChanges = true;
    this._root = Anyfile.createFromJSON(this, {
        filename: ((typeof json.rootName == 'string') || (json.rootName instanceof String)) ? json.rootName : '',
        data: (json.files instanceof Array) ? json.files : []
    });
    if (json.clipboard instanceof Object) {
        this.boolMove = json.clipboard.move ? json.clipboard.move : false;
        this._clipboard = json.clipboard.path ? this._root.getByPath(json.clipboard.path) : null;
    } else {
        this.boolMove = false;
        this._clipboard = null;
    }
    this._current = this._root;
    this.path = ((typeof json.path == 'string') || (json.path instanceof String)) ? json.path : '/';
    this._ignoreChanges = undefined;
    this.reportChange(this._root);
}

/**
* Aktuellen Status und Inhalt in JSON umwandeln
* @return Object    JSON
*/
FileManager.prototype.toJSON = function() {
    var root = this._root.toJSON();
    var o = { path: this.path, files: root.data };
    if (root.filename != '')
        o.rootName = root.filename;
    if (this.useClipboard && this._clipboard)
        o.clipboard = { path: this._clipboard.getPath(), move: this.boolMove };
    return o;
}

/**
* Angezeigte Dateiliste aktualisieren
* @return undefined
*/
FileManager.prototype.refresh = function() {
if (!(this._current instanceof Anyfile)) return;
    var z, index, className;
    var arrFiles = this._current.data.slice(0);
    // Zu viel angezeigte Dateien entfernen und alte Dateianzeigen anpassen
    for (z = 0; z < this._arrLis.length; ++z) {
        index = arrFiles.indexOf(this._arrLis[z].file);
        if (index == -1) {
            this._nodeUl.removeChild(this._arrLis[z].node);
            this._arrLis.splice(z, 1);
            --z;
        } else {
            if (this._arrLis[z].node.firstChild.textContent != arrFiles[index].filename)
                this._arrLis[z].node.firstChild.textContent = arrFiles[index].filename;
            className = FileManager.getFileClass(arrFiles[index]);
            if (this._arrLis[z].node.firstChild.className != className)
                this._arrLis[z].node.firstChild.className = className;
            className = (this.boolMove && this.useClipboard && (this._arrLis[z].file == this._clipboard)) ? 'cut' : '';
            if (this._arrLis[z].node.className != className)
                this._arrLis[z].node.className = className;
            arrFiles.splice(index, 1);
        }
    }
    // Neue Dateianzeigen erstellen
    while(arrFiles.length)
        this._showFile(arrFiles.pop());
    // Sortieren
    if (!this._arrLis.length)
        return;
    var boolDirsFirst = this._boolDirsFirst;
    this._arrLis.sort(function(a, b) {
        if (boolDirsFirst) {
            var boolDirA = a.file.isDir;
            var boolDirB = b.file.isDir;
            if (boolDirA != boolDirB)
                return (boolDirA ? -1 : +1);
        }
        return a.file.filename.localeCompare(b.file.filename);
    });
    // Reihenfolge ggf. anpassen
    if (this._nodeUl.firstChild != this._arrLis[0].node)
        this._nodeUl.insertBefore(this._arrLis[0].node, this._nodeUl.firstChild);
    for (z = 1; z < this._arrLis.length; ++z) {
        if (this._nodeUl.childNodes[z] != this._arrLis[z].node)
            this._nodeUl.insertBefore(this._arrLis[z].node, this._nodeUl.childNodes[z]);
    }
}

/**
* Überprüft, ob der aktuelle Zustand alle Kriterien erfüllt
* @return Object    Erste gefundene Abweichung
*/
FileManager.prototype.check = function(
    arrTests    // Array aus Objekten der Form { path: "Absoluter Dateipfad", isDir: "Handelt es sich um ein Verzeichnis (optional)", data: "Dateiinhalt (optional)", type: "Dateityp (optional)", contains: "Anzahl an Dateien im Verzeichnis (optional)" }
) {
    var file;
    for (var z = 0; z < arrTests.length; ++z) {
        file = this._root.getByPath(arrTests[z].path);
        if (!file)
            return { type: 'Missing file', path: arrTests[z].path, objTest: arrTests[z] };
        if (arrTests[z].data) {
            if (file.data != arrTests[z].data)
                return { type: 'Wrong file content', path: arrTests[z].path, objTest: arrTests[z] };
        }
        if (((arrTests[z].isDir !== undefined) && (file.isDir != arrTests[z].isDir)) || (arrTests[z].contains !== undefined) && !file.isDir)
            return { type: arrTests[z].isDir ? 'No directory' : 'No file', path: arrTests[z].path, objTest: arrTests[z] };
        if ((arrTests[z].contains !== undefined) && (file.data.length != arrTests[z].contains))
            return { type: 'Wrong number of subfiles', path: arrTests[z].path, objTest: arrTests[z] };
        if (arrTests[z].type && (file.type != arrTests[z].type))
            return { type: 'Wrong type', path: arrTests[z].path, objTest: arrTests[z] };
    }
    return null;
}
